/*
Module : MfcCDDB.cpp
Purpose: Defines the implementation for a collection of MFC classes to wrap access to CDDB
Created: PJN / 24-02-1999
History: PJN / 19-05-1999 1. Added support for MOTD (Message of the Day)
                          2. Added support for submit
                          3. Fixed a number of places where calling the TRACE
                          function was causing an access violation
                          4. Fixed a bug in GetErrorMessage which was causing
                          SDK errors to return the string "The operation completed successfully"
         PJN / 14-07-1999 1. Fixed two potential access violations in the CCDDB::ReadResponse function.
         PJN / 29-11-1999 Fixed a number of bugs when compiled with VC 6.0
         PJN / 17-01-2000 Fixed a number of compiler errors which seem to have crept into v1.12
         PJN / 23-01-2000 1. The code has now been fully tested with the open source alternative to CDDB namely 
                          freedb.freedb.org. CDDB has become more commercialized of late, since its introduction 
                          of a binary SDK with closed validation and new licensing rules. I also believe that 
                          CDDB will stop access sometime in 2000 to its databases via non CDDB originated code.
                          Therefore I would suggest that developers using MfcCDDB should provide users with a 
                          configuration dialog to allow the CDDB server to be specified, that way you will not 
                          have irate customers ringing you when CDDB stops allowing access to it. You can use 
                          this open source server instead of the CDDB one in the MfcCDDB code by using 
                          "freedb.freedb.org" instead of "cddb.cddb.com" in the call to CCDDB::Sites and by 
                          specifying CDDB protocol version 3 in the function SetCDDBProtocolVersion.
         PJN / 13-02-2000 1. Now fully supports access when behind a HTTP proxy.
                          2. Also now includes support for proxy authentication
         PJN / 07-05-2000 Fixed a problem in GetUserName and GetHostName with embedded spaces
         PJN / 17-05-2000 Now handles CDDB records which have more than one DTITLE line
         PJN / 30-05-2000 Now prevents creation of empty extended track tags (EXTT=)
         PJN / 05-09-2000 Modified CCDDB::GetTrackPositions to fail when the CD does not contain any audio tracks
         PJN / 09-08-2001 Fixed a bug where the first track title was not being parsed correctly. Thanks to Jared 
                          Allen for spotting this.
         PJN / 06-10-2001 Major update to bring code into line with freedb standard. As mentioned in previous 
                          updates you can no longer use MFCCDDB with the commercial CDDB web site at Gracenote
                          (formerly CDDB). Instead now the code can only be used with the open source alternative
                          of freedb. Updates include:
                          1. Updated copyright information
                          2. Fixed calculation the extra disc length parameter, which must be specified in the query string.
                          3. Fixed bug where discid for submissions was having any leading 0's stripped.
                          4. Updated m_sProductVersion as it was out of date.
                          5. Changed default server in Sites method to "freedb.freedb.org"
                          6. Removed comment line which says "Copyright ... CDDB Inc" from Submit method
                          7. Updated links in documentation to be www.freedb.org
                          8. Fixed a bug in ReadResponse when an error occurs reading from the socket
                          9. Reworked the ReadResponse method to be more inline with my W3MFC socket class
                          10. Updated the sample app to display the last command response when a function fails
                          11. framework now works with version 5 of the CDDB protocol. This gives two extra
                          fields in the CDDB record class namely Year and Genre.
                          12. Complete rewrite of the ParseRecordBody to improve readability of the function. This 
                          means that the code now correctly supports all text record values which span multiple 
                          lines
                          13. Made m_ExtendedData member of CCDDBRecord a CString instead of a string array
                          14. Fixed a bug in CCDDB::ParseKeywordLine
                          15. Fixed bug when separating artist and title from DTITLE lines
                          16. Fixed an infinite recursion bug in ReplaceSubString
         PJN / 02-02-2002 1. Fixed a minor parsing bug in CCDDB::GetNextLine. Thanks to Thomas Lang for this fix.
         PJN / 24-06-2002 1. Now fully supports local access to database. The standard and alternate form are
                          both fully supported.
                          2. Optimized #include of 2 files in mfccddb.cpp
         PJN / 03-08-2003 1. Fixed a bug when the MCI device was not being closed if the inserted CD did not contain
                          an Audio CD. Thanks to Christian Inkster for reporting this problem.
         PJN / 03-11-2003 1. Simplified the code in CCDDB::ReadResponse. Thanks to Clarke Brunt for reporting this issue.
         PJN / 20-08-2004 1. General tidy up of the GetUserName and GetHostName functions.
         PJN / 29-06-2005 1. Updated copyright details.
                          2. Updated comments regarding how to include headers into your pre compiled header
                          3. Reviewed all TRACE statements for correctness
                          4. Updates to the GetUserName and GetHostName functions to exclude high ASCII characters.
                          Thanks to Stefan Becker for providing the fix for this bug.
                          5. Removed unreferenced variable warning in CCDDB::FormSubmitLine
                          6. Fixed a number of level 4 warnings when code is compiled in Visual Studio .NET 2003
                          7. Fixed various warnings when code is compiled with /Wp64 compiler setting in
                          Visual Studio .NET 2003. Please note that to achieve this, the code now requires a recentish
                          Platform SDK to be installed and configured if the code is compiled on Visual C++ 6
         PJN / 01-01-2007 1. Updated copyright details.
                          2. Addition of a MFCCDDB_EXT_CLASS preprocessor macro to allow the classes to be more easily 
                          added to an extension dll
                          3. Updated the code to clean compile in VC 2005
                          4. Updated documentation to use the same style as the web site
                          5. Default timeout is now 60 seconds for release builds and 90 seconds for debug builds
                          6. Reviewed all CString::Format calls for correctness
                          7. Optimized CCDDBSite constructor code
                          8. Optimized CCDDBQueryResult constructor code
                          9. Optimized CCDDBTrackPosition constructor code
                          10. Optimized CCDDBRecord constructor code
                          11. Optimized CCDDB constructor code
                          12. Removed unused CCDDB destructor
                          13. Fixed Unicode issue in CCDDB::GetCDROMDrives
                          14. Updated usage of CString::TrimLeft and CString::TrimRight in the code
                          15. Replaced usage of CCDDB::ReplaceSubString with CString::Replace
                          16. Hostname can now be customized via a SetHostName method
                          17. m_sLastCommandResponse is now not reset when a response is successfully parsed
                          18. UserName can now be customized via a SetUserName method
                          19. CCDDBSite::Parse now does not fail if protocol is not "http"
                          20. Reworked the classes to be exception based along the lines of the author's CPJNSMTPConnection
                          class. This should improve the robustness of client application code. Please note that because
                          of this redesign, you should reexamine all client code to make sure it catches the new
                          exceptions thrown from the classes as appropriate
                          21. Fixed a bug in CCDDB::SplitSemiColonLine where 1 character length values would not be 
                          parsed correctly
                          22. Added a constructor to the class CCDDBStatus
                          23. Fixed a bug in CCDDB::ParseCommentLine where 1 character length values would not be
                          parsed correctly
                          24. Fixed a bug in CCDDB::ReadResponse where it had problems parsing the body of the response
                          for query and submit requests if the body arrives in a second packet from the network
                          25. CCDDB::ParseRecordBodyFromFile now allows other processes to read the file which it is
                          currently attempting to parse
                          26. Optimized code in CCDDB::Read(const CString& sDirectory... which operates on the passed in
                          dwDiscID parameter
                          27. Optimized code in CCDDB::ParseRecordBodyFromFile to handle parsing the non-standard (Windows)
                          format of CDDB record format
         PJN / 12-01-2008 1. Updated copyright details.
                          2. Updated sample app code to remove VC 6 style classwizard comments
                          3. Updated the sample app to clean compile on VC 2008
                          4. CCDDBException::GetErrorMessage now uses the FORMAT_MESSAGE_IGNORE_INSERTS flag. For more 
                          information please see Raymond Chen's blog at 
                          http://blogs.msdn.com/oldnewthing/archive/2007/11/28/6564257.aspx. Thanks to Alexey Kuznetsov for 
                          reporting this issue.
                          5. CCDDBException::GetErrorMessage now uses Checked::tcsncpy_s if compiled using VC 2005 or 
                          later.
         PJN / 31-08-2008 1. Code now compiles cleanly using Code Analysis (/analyze)
                          2. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          3. The code now only supports VC 2005 or later. 
                          4. Added support for parsing CCDDBSite::m_sProtocol
                          5. Code now uses ATL's support for base64 encoding
                          6. The code now uses the author's CWSocket to encapsulate socket functionality
                          7. Fixed a bug in the calculation of the total track length in CCDDB::ComputeDiscID. Thanks
                          to Tom van Dijck for reporting this bug.
                          8. General tidy up of all the parsing code in the classes
                          9. The classes now supports v6 Protocol level (if the code is compiled for Unicode)
                          10. GetTrackPositions has been renamed to GetTrackPositionsUsingMCI.
                          11. Addition of a GetTrackPositionsUsingDeviceIoControl using the API DeviceIoControl and the 
                          IOCTL code of IOCTL_CDROM_READ_TOC to get the track positions. This method works on all 
                          versions of Windows based on the NT Kernel (i.e. NT, 2000, XP, 2003, Vista and 2008). It
                          should also work for multi-session CD's i.e. CD's which have data sessions as well as an
                          audio session. 
                          12. ComputeDiscID has been renamed to ComputeDiscIDUsingMCI.
                          13. Addition of a ComputeDiscIDUsingDeviceIoControl method.
         PJN / 24-12-2009 1. Removed various preprocessor defines from top of MfcCDDB as they are available in standard
                          VC headers.
                          2. Fixed up a minor case sensitively issue with the various "MfcCDBD" files.
                          3. Updated copyright details.
                          4. Updated the sample app configuration settings to more modern default values.
                          5. Fix a bug where the drive/device name created in GetTrackPositionsUsingDeviceIoControl was 
                          incorrectly created as "\\.\\[Drive Letter]:" instead of the correct format of 
                          "\\.\[Drive Letter]:". This could lead to this function call failing in certain circumstances.
                          Thanks to "Jase" for reporting this issue.
                          6. Updated the code in CCDDB::ReadResponse in line with the similar response parsing code in the
                          CPJNPOP3Connection v1.47, This helps address an issue where the terminator could not be found if
                          the response gets split across socket calls. This update also helps avoid excessive CPU utilization 
                          when we receive large responses. Thanks to Markus Hlacer, "Jase" and Alexey Rabets for reporting 
                          this issue.
                          7. Eliminated use of old style CT2A / CA2T macros and replaced with CStringA/W.
         PJN / 10-04-2010 1. Updated copyright details
                          2. CCDDB::ParseQueryBody now does not consider receiving a 202 response as an error
         PJN / 26-02-2016 1. Updated copyright details.
                          2. Removed all the proxy connection methods as they cannot be easily supported / tested by the 
                          author.
                          3. Updated the code to work with the latest version of the author's CWSocket classes. The CWSocket
                          class requires VC 2008 at a minimum to compile now and this is the earliest supported version of
                          VC for the MfcCDDB classes now.
                          4. Updated the code to clean compile on VC 2010 - VC 2015
                          5. Reworked ParseRecordBodyFromFile method to use ATL::CHeapPtr class.
                          6. Reworked the MapFromProtocolString & MapToProtocolString methods internal implementation.
                          7. All the classes have been moved to a C++ namespace called "CDDB"
                          8. Renamed the "CCDDB" class to "CClient".
                          9. Removed the Read(const String& sDirectory...) method as this requires a full copy of the 
                          freedb database on the local file system and this is now approaching 1 GB in size as of Q1 2016.
                          10. Removed the ParseRecordBodyFromFile method.
                          11. Reworked the classes to optionally compile without MFC. By default the class now use STL 
                          classes and idioms but if you define CWSOCKET_MFC_EXTENSTIONS the class will revert back to the 
                          MFC behaviour.
                          12. Added SAL annotations to all the code.
                          13. Changed ordering of parameters in GetTrackPositionsUsingMCI,  
                          GetTrackPositionsUsingDeviceIoControl, ParseRecordBodyHelper, ParseSiteBodyHelper,
                          ParseQueryBodyHelper & Sites methods.
                          14. ParseRecordBodyHelper, ParseSiteBodyHelper, ParseQueryBodyHelper and ParseStatusBody now 
                          calls SetLastError with ERROR_INVALID_DATA if they detects parsing errors
         PJN / 30-07-2016 1. Updated code to use new preprocessor define of CWSOCKET_MFC_EXTENSIONS instead of 
                          CWSOCKET_MFC_EXTENSTIONS
         PJN / 20-05-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly using /permissive-.
         PJN / 04-10-2017 1. Replaced NULL throughout the codebase with nullptr. This means that the minimum
                          requirement for the framework is now VC 2010.
                          2. Replaced CString::operator LPC*STR() calls throughout the codebase with
                          CString::GetString calls
         PJN / 26-12-2017 1. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is defined.
         PJN / 02-01-2019 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes
                          mean that the code will now only compile on VC 2017 or later.
                          3. Removed code which supported CWSOCKET_MFC_EXTENSIONS preprocessor define
         PJN / 06-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 29-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 23-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 12-04-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 28-12-2021 1. Updated copyright details.
                          2. The code has now been tested with gnudb.gnudb.org now that freedb.org has been closed.
                          3. Updated links in documentation to be gnudb.org
         PJN / 28-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.
         PJN / 14-08-2022 1. Reworked network code to use author's WinHTTPWrappers classes instead of raw sockets
                          to connect via HTTP to the CDDB servers.
         PJN / 17-08-2022 1. Reworked to use latest version of WinHTTPWrappers classes

Copyright (c) 1999 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#ifndef _INC_MMSYSTEM
#pragma message("To avoid this message, please put mmsystem.h in your pre compiled header (normally stdafx.h)")
#include <mmsystem.h>
#endif //#ifndef _INC_MMSYSTEM
#ifndef NETCONS_INCLUDED
#pragma message("To avoid this message, please put lmcons.h in your pre compiled header (normally stdafx.h)")
#include <lmcons.h>
#endif //#ifndef NETCONS_INCLUDED
#ifndef __ATLBASE_H__
#pragma message("To avoid this message, please put atlbase.h in your pre compiled header (normally stdafx.h)")
#include <atlbase.h>
#endif //#ifndef __ATLBASE_H__
#include <WinIoCtl.h>
#ifndef _NTDDCDRM_
#pragma message("To avoid this message, please put ntddcdrm.h in your pre compiled header (normally stdafx.h)")
#include <ntddcdrm.h>
#endif //#ifndef _NTDDCDRM_
#ifndef _SSTREAM_
#pragma message("To avoid this message, please put sstream in your pre compiled header (normally stdafx.h)")
#include <sstream>
#endif //#ifndef _SSTREAM_
#ifndef _IOMANIP_
#pragma message("To avoid this message, please put iomanip in your pre compiled header (normally stdafx.h)")
#include <iomanip>
#endif //#ifndef _IOMANIP_
#ifndef _IOS_
#pragma message("To avoid this message, please put ios in your pre compiled header (normally stdafx.h)")
#include <ios>
#endif //#ifndef _IOS_
#include "MfcCDDB.h"
#include "resource.h"


//////////////////// Statics / Macros /////////////////////////////////////////

//Automatically pull in the library winmm.lib since we
//need MCI to retrieve the cdaudio track details
#pragma comment(lib, "winmm.lib")

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

CDDB::CException::CException(_In_ HRESULT hr, _In_z_ LPCSTR pszLastResponse) : m_bMMError{false},
                                                                               m_hr{hr},
                                                                               m_sLastResponse{pszLastResponse}
{
}

CDDB::CException::CException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_z_ LPCSTR pszLastResponse) : m_bMMError{false},
                                                                                                         m_hr{MAKE_HRESULT(SEVERITY_ERROR, dwFacility, dwError)},
                                                                                                         m_sLastResponse{pszLastResponse}
{
}

CDDB::CException::CException(_In_ bool bMMError, _In_ DWORD dwError, _In_z_ LPCSTR pszLastResponse) : m_bMMError{bMMError},
                                                                                                      m_hr{MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, dwError)},
                                                                                                      m_sLastResponse{pszLastResponse}
{
}

#pragma warning(suppress: 26429)
bool CDDB::CException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //If would be great if there was a FACILITY_MMSYSTEM, but there isn't, so we carry around an extra boolean
  //in the exception class which signifies if it is a MMSYSTEM error
  if (m_bMMError)
  {
    bSuccess = mciGetErrorString(HRESULT_CODE(m_hr), lpszError, nMaxError);
    if (!bSuccess)
      *lpszError = _T('\0');
  }
  else if (HRESULT_FACILITY(m_hr) == FACILITY_ITF)
  {
    bSuccess = true;

    //Simply load up the string from the string table
    String sError;
    sError.resize(4096);
#pragma warning(suppress: 26446 26472)
    LoadString(GetModuleHandle(nullptr), HRESULT_CODE(m_hr), sError.data(), static_cast<int>(sError.size()));
    const auto nFind{sError.find(_T("%1"))};
    if (nFind != String::npos)
    {
    #ifdef _UNICODE
      sError.replace(nFind, 2, ATL::CA2W(m_sLastResponse.c_str()).operator LPWSTR());
    #else
      sError.replace(nFind, 2, m_sLastResponse);
    #endif //#ifdef _UNICODE
    }
    ATL::Checked::tcsncpy_s(lpszError, nMaxError, sError.c_str(), _TRUNCATE);
  }
  else
  {
    LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
    const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                       nullptr, HRESULT_CODE(m_hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
                                       reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};

    if (dwReturn == 0)
      *lpszError = _T('\0');
    else
    {
      bSuccess = true;
      ATL::Checked::tcsncpy_s(lpszError, nMaxError, lpBuffer, _TRUNCATE);
      LocalFree(lpBuffer);
    }
  }

  return bSuccess;
}


#pragma warning(suppress: 26455)
CDDB::CSite::CSite(_In_ bool bRetrieve) : m_nPort{80},
                                          m_bNorthing{true},
                                          m_nLatitudeMinutes{0},
                                          m_bEasting{true},
                                          m_nLongitudeMinutes{0}
{
  if (bRetrieve)
    m_sAddress = _T("/~cddb/cddb.cgi");
  else
    m_sAddress = _T("/~cddb/submit.cgi");
}


#pragma warning(suppress: 26455)
CDDB::CClient::CClient() : m_sProductName{_T("MfcCDDB")},
                           m_sProductVersion{_T("1.47")},
                         #ifdef _UNICODE
                           m_nProtocolVersion{6}
                         #else
                           m_nProtocolVersion{5}
                         #endif //#ifdef _UNICODE
{
  SetHostname(_T("auto"));
  SetUserName(_T("auto"));
}

void CDDB::CClient::SetCDDBProtocolVersion(_In_ int nProtocolVersion) noexcept
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(nProtocolVersion >= 3 && nProtocolVersion <= 6); //MfcCDDB supports v3, v4, v5 & v6 only.

  m_nProtocolVersion = nProtocolVersion;
}

#pragma warning(suppress: 26440)
DWORD CDDB::CClient::ComputeDiscID(_In_ const std::vector<CTrackPosition>& tracks)
{
  //Number of tracks is 1 less than the size of the array as it also contains the lead out position
#pragma warning(suppress: 26472)
  const auto nTracks{static_cast<int>(tracks.size() - 1)};

  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(nTracks > 0);

  //Iterate across all the tracks
  int n{0};
  for (int i{0}; i<nTracks; i++)
  {
    int sum{0};
#pragma warning(suppress: 26446)
    int j{tracks[i].m_nMinute*60 + tracks[i].m_nSecond};
    while (j > 0)
    {
      sum += (j % 10);
      j /= 10;
    }
    n += sum;
  }

  //Compute total track length in seconds
#pragma warning(suppress: 26446)
  const int nTotalTrackLength{tracks[nTracks].m_nMinute*60 + tracks[nTracks].m_nSecond - tracks[0].m_nMinute*60 - tracks[0].m_nSecond};

  //Compute DISC ID
#pragma warning(suppress: 26472)
  const auto dwDiscID{static_cast<DWORD>((n % 0xFF) << 24 | (nTotalTrackLength << 8) | nTracks)};
  return dwDiscID;
}

void CDDB::CClient::GetTrackPositionsUsingDeviceIoControl(_In_ LPCTSTR pszDrive, _Out_ std::vector<CTrackPosition>& tracks)
{
  //Remove any tracks already in the array
  tracks.clear();

  //Convert the drive format layout of the "pszDrive" parameter to the device format
  String sDevice{_T("\\\\.\\")};
  sDevice += pszDrive;

  //first open the device
  ATL::CHandle handle{CreateFile(sDevice.c_str(), GENERIC_READ, FILE_SHARE_READ, nullptr, OPEN_EXISTING, 0, nullptr)};
  if (handle == INVALID_HANDLE_VALUE)
    ThrowCDDBException(GetLastError(), FACILITY_WIN32);

  //Issue the IOCTL to read the toc
  CDROM_TOC toc{};
  DWORD dwBytesReturn{0};
  if (!DeviceIoControl(handle, IOCTL_CDROM_READ_TOC, nullptr, 0, &toc, sizeof(toc), &dwBytesReturn, nullptr))
    ThrowCDDBException(GetLastError(), FACILITY_WIN32);

  //If successful, fill in the output parameter array
  const size_t nTotalTracks{static_cast<size_t>(toc.LastTrack) - toc.FirstTrack + 1};
  tracks.resize(nTotalTracks + 1);
  for (UCHAR nTrack{toc.FirstTrack}; nTrack<=(toc.LastTrack+1); nTrack++) //Note this code treats filling in the lead out position just like any other track
  {
    //Save the track position in MSF format
    CTrackPosition trackPosition;
#pragma warning(suppress: 26446 26482)
    trackPosition.m_nMinute = toc.TrackData[nTrack-toc.FirstTrack].Address[0]*60 + toc.TrackData[nTrack-toc.FirstTrack].Address[1];
#pragma warning(suppress: 26446 26482)
    trackPosition.m_nSecond = toc.TrackData[nTrack-toc.FirstTrack].Address[2];
#pragma warning(suppress: 26446 26482)
    trackPosition.m_nFrame  = toc.TrackData[nTrack-toc.FirstTrack].Address[3];

    //Store the value in the array
#pragma warning(suppress: 26446 26489)
    tracks[nTrack - toc.FirstTrack] = trackPosition;
  }
}

void CDDB::CClient::GetTrackPositionsUsingMCI(_In_ LPCTSTR pszDrive, _Out_ std::vector<CTrackPosition>& tracks)
{
  //Remove any tracks already in the array
  tracks.clear();

  //Open the specified "cdaudio" MCI device
  MCI_OPEN_PARMS mciOpenParms;
  mciOpenParms.lpstrDeviceType = _T("cdaudio");
  mciOpenParms.lpstrElementName = pszDrive;
#pragma warning(suppress: 26490)
  DWORD dwError{mciSendCommand(0, MCI_OPEN, MCI_OPEN_SHAREABLE | MCI_OPEN_TYPE | (pszDrive ? MCI_OPEN_ELEMENT : 0), reinterpret_cast<DWORD_PTR>(&mciOpenParms))};
  if (dwError)
    ThrowCDDBMMException(dwError);

  //Set the time format to Minute/Second/Frame (MSF) format
  MCI_SET_PARMS mciSetParms;
  mciSetParms.dwTimeFormat = MCI_FORMAT_MSF;
#pragma warning(suppress: 26490)
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_SET, MCI_SET_TIME_FORMAT, reinterpret_cast<DWORD_PTR>(&mciSetParms));
  if (dwError)
  {
    //Don't forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }

  //Get the total track count
  MCI_STATUS_PARMS mciStatusParms;
  mciStatusParms.dwItem = MCI_STATUS_NUMBER_OF_TRACKS;
#pragma warning(suppress: 26490)
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
  if (dwError)
  {
    //Don't forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }
#pragma warning(suppress: 26472)
  const auto nTotalTracks{static_cast<size_t>(mciStatusParms.dwReturn)};

  //check that disk contain audio track
  bool bAudioTrackPresent{false};
  for (size_t i{1}; i<=nTotalTracks && !bAudioTrackPresent; i++)
  {
    mciStatusParms.dwItem = MCI_CDA_STATUS_TYPE_TRACK;
#pragma warning(suppress: 26472)
    mciStatusParms.dwTrack = static_cast<DWORD>(i);
#pragma warning(suppress: 26490)
    dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
    if (dwError)
    {
      //Don't forget to close the MCI device before we throw the exception
      mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

      ThrowCDDBMMException(dwError);
    }
    if (mciStatusParms.dwReturn == MCI_CDA_TRACK_AUDIO)
      bAudioTrackPresent = true;
  }
  if (!bAudioTrackPresent)
  {
    //Don't forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBException(IDS_MFCCDDB_NO_AUDIO_TRACKS_PRESENT, FACILITY_ITF);
  }

  //Iterate through all the tracks getting their starting position
  tracks.resize(nTotalTracks + 1);
  for (size_t i{1}; i<=nTotalTracks; i++)
  {
    mciStatusParms.dwItem = MCI_STATUS_POSITION;
#pragma warning(suppress: 26472)
    mciStatusParms.dwTrack = static_cast<DWORD>(i);
#pragma warning(suppress: 26490)
    dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
    if (dwError)
    {
      //Don't forget to close the MCI device before we throw the exception
      mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

      ThrowCDDBMMException(dwError);
    }

    //Save the track position in MSF format
    CTrackPosition trackPosition;
    trackPosition.m_nMinute = MCI_MSF_MINUTE(mciStatusParms.dwReturn);
    trackPosition.m_nSecond = MCI_MSF_SECOND(mciStatusParms.dwReturn);
    trackPosition.m_nFrame  = MCI_MSF_FRAME(mciStatusParms.dwReturn);

    //Store the value in the array
#pragma warning(suppress: 26446 26489)
    tracks[i-1] = trackPosition;
  }

  //Get the last track's length
  mciStatusParms.dwItem = MCI_STATUS_LENGTH;
#pragma warning(suppress: 26472)
  mciStatusParms.dwTrack = static_cast<DWORD>(nTotalTracks);
#pragma warning(suppress: 26490)
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
  if (dwError)
  {
    //Don't forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }

  //Compute lead-out track position
  const DWORD dwLenM{MCI_MSF_MINUTE(mciStatusParms.dwReturn)};
  const DWORD dwLenS{MCI_MSF_SECOND(mciStatusParms.dwReturn)};
#pragma warning(suppress: 26472)
  const auto dwLenF{static_cast<DWORD>(MCI_MSF_FRAME(mciStatusParms.dwReturn) + 1)}; //Fix MCI Windows bug according to CDDB Howto doc
#pragma warning(suppress: 26446 26472 26489)
  const auto dwPosM{static_cast<DWORD>(tracks[nTotalTracks-1].m_nMinute)};
#pragma warning(suppress: 26446 26472 26489)
  const auto dwPosS{static_cast<DWORD>(tracks[nTotalTracks - 1].m_nSecond)};
#pragma warning(suppress: 26446 26472 26489)
  const auto dwPosF{static_cast<DWORD>(tracks[nTotalTracks-1].m_nFrame)};

  //Compute lead out track position (in frame format)
  DWORD dwPos{(dwPosM*60*75) + (dwPosS*75) + dwPosF + (dwLenM*60*75) + (dwLenS*75) + dwLenF};

  //Convert dwPos back to MSF format
  CTrackPosition trackPosition;
  trackPosition.m_nFrame = dwPos % 75;
  dwPos /= 75;
  trackPosition.m_nSecond = dwPos % 60;
  dwPos /= 60;
  trackPosition.m_nMinute = dwPos;

  //And store in the array
#pragma warning(suppress: 26446 26489)
  tracks[nTotalTracks] = trackPosition;

  //Don't forget to close the MCI device before we return
  mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);
}

DWORD CDDB::CClient::ComputeDiscIDUsingMCI(_In_ LPCTSTR pszDrive)
{
  //Get the track details
  std::vector<CTrackPosition> tracks;
  GetTrackPositionsUsingMCI(pszDrive, tracks);

  //Compute the DISC ID now that we have got all the track information
  return ComputeDiscID(tracks);
}

DWORD CDDB::CClient::ComputeDiscIDUsingDeviceIoControl(_In_ LPCTSTR pszDrive)
{
  //Get the track details
  std::vector<CTrackPosition> tracks;
  GetTrackPositionsUsingDeviceIoControl(pszDrive, tracks);

  //Compute the DISC ID now that we have got all the track information
  return ComputeDiscID(tracks);
}

void CDDB::CClient::GetCDROMDrives(_Out_ std::vector<String>& drives)
{
  //empty out the array
  drives.clear();

  //Iterate across all the drive letters to find out which ones are CDROMs
  for (int i{1}; i<=26; i++)
  {
    String sDrive;
#pragma warning(suppress: 26472)
    sDrive += static_cast<TCHAR>((_T('A') + i - 1));
    sDrive += _T(":");
    if (GetDriveType(sDrive.c_str()) == DRIVE_CDROM)
#pragma warning(suppress: 26489)
      drives.push_back(sDrive);
  }
}

#pragma warning(suppress: 26440)
void CDDB::CClient::StripNonAlphaOrDigits(_Inout_ String& sValue)
{
  const auto nLength{sValue.length()};
  for (String::size_type i{0}; i<nLength; i++)
  {
#pragma warning(suppress: 26446 26489)
    const TCHAR c{sValue[i]};
    const bool bIsAlpha{(((c >= _T('A')) && (c <= _T('Z'))) || //Definitively between A-Z or a-z and nothing else!
                        ((c >= _T('a')) && (c <= _T('z'))))};
    const bool bIsDigit{((c >= _T('0')) && (c <= _T('9')))};
    if (!(bIsAlpha || bIsDigit))
#pragma warning(suppress: 26446 26489)
      sValue[i] = _T('_');
  }
}

void CDDB::CClient::SetUserName(_In_ const String& sUserName)
{
  if (sUserName == _T("auto"))
  {
    //Get the current windows user name
    String sUser;
    sUser.resize(UNLEN + 1);
    DWORD dwSize{UNLEN + 1};
#pragma warning(suppress: 26446)
    const BOOL bGotUserName{::GetUserName(sUser.data(), &dwSize)};
    sUser.resize(dwSize - 1);
    if (bGotUserName)
      m_sUserName = sUser;
    else
      m_sUserName = _T("Unknown");
  }
  else
    m_sUserName = sUserName;

  //convert string to alpha chars and digits only
  StripNonAlphaOrDigits(m_sUserName);
}

void CDDB::CClient::SetHostname(_In_ const String& sHostname)
{
  //retrieve the localhost name if we are supplied the magic "auto" host name
  if (sHostname == _T("auto"))
  {
    //get the current windows host name
    StringA sHostName;
    sHostName.resize(256);
#pragma warning(suppress: 26446)
    gethostname(sHostName.data(), 256);
    m_sHostName = ATL::CA2T{sHostName.c_str()};
  }
  else
    m_sHostName = sHostname;

  //convert string to alpha chars and digits only
  StripNonAlphaOrDigits(m_sHostName);
}

CDDB::String CDDB::CClient::GetHelloCommand()
{
  String sCommand;
#ifdef _UNICODE
  std::wostringstream ss;
#else
  std::ostringstream ss;
#endif //#ifdef _UNICODE
  ss << _T("hello=") << GetUserName() << _T("+") << GetHostName() << _T("+") << m_sProductName << _T("+") << m_sProductVersion << _T("&proto=") << m_nProtocolVersion;
  sCommand = ss.str();
  return sCommand;
}

int CDDB::CClient::GetCDDBResponseCode(_In_ LPCSTR pszBody)
{
  //What will be the return value from this function
  int nResponse{-1};

  if (pszBody != nullptr)
  {
    const auto nLength{strlen(pszBody)};
    if (nLength >= 3)
    {
      StringA sNum{pszBody, 3};
      nResponse = atoi(sNum.c_str());
    }
  }

  return nResponse;
}

_Return_type_success_(return != 0) bool CDDB::CClient::SplitSemiColonLine(_In_ const StringA& sLine, _Out_ StringA& sField, _Out_ StringA& sValue)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Do the actual parsing
  const auto nSeparator{sLine.find(":")};
  if (nSeparator != StringA::npos)
  {
    sField = sLine.substr(0, nSeparator);
    sValue = sLine.substr(nSeparator + 1, sLine.length() - nSeparator - 1);

    //Remove unwanted spaces
    StringTrim(sField);
    StringTrim(sValue);

    bSuccess = true;
  }

  return bSuccess;
}

void CDDB::CClient::Sites(_In_ const CDDB::CSite& server, _Out_ std::vector<CSite>& sites)
{
  //Initialize the output parameter
  sites.clear();

  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request
  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=sites&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse the sites info from the response
  if (!ParseSitesBody(sResponse.c_str(), sites))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SITES_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseSitesBody(_In_ LPCSTR pszBody, _Out_ std::vector<CSite>& sites)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Initialize the output parameter
  sites.clear();

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then parse the site data line by line
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    //Skip over the current line
    LPCSTR pszLine{pszBody};
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line
    while (pszLine != nullptr)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //If we can parse the current line then add it 
      //to the sites array
      if (sLine != ".")
      {
        CSite site;
#pragma warning(suppress: 26486)
        if (ParseSiteBodyHelper(sLine, site))
#pragma warning(suppress: 26489)
          sites.push_back(site);
      }
    }

    bSuccess = true;
  }

  return bSuccess;
}

void CDDB::CClient::Sites(_Out_ std::vector<CSite>& sites, _In_ const String& sServer, _In_ const String& sAddress, _In_ INTERNET_PORT nPort)
{
  //Form the site which we will be querying
  CSite server;
#pragma warning(suppress: 26489)
  server.m_sSite = sServer;
  server.m_nPort = nPort;
#pragma warning(suppress: 26489)
  server.m_sAddress = sAddress;

  //Let the other version do the heavy lifting
#pragma warning(suppress: 26486)
  Sites(server, sites);
}

void CDDB::CClient::Categories(_In_ const CDDB::CSite& server, _Out_ std::vector<String>& categories)
{
  //Initialize the output parameter
  categories.clear();

  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request
  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=cddb+lscat&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse the categories info from the response
  if (!ParseCategoriesBody(sResponse.c_str(), categories))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_CATEGORIES_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseCategoriesBody(_In_ LPCSTR pszBody, _Out_ std::vector<String>& categories)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Initialize the output parameter
  categories.clear();

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then parse the categories data line by line
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    //Skip over the current line
    LPCSTR pszLine{pszBody};
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line
    while (pszLine != nullptr)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //Add the current line to the categories array
      if (sLine.length() && sLine != ".")
#pragma warning(suppress: 26489)
        categories.push_back(MapFromProtocolString(sLine));
    }

    bSuccess = true;
  }

  return bSuccess;
}

void CDDB::CClient::Status(_In_ const CDDB::CSite& server, _Out_ CDDB::CStatus& status)
{
  //Initialize the output parameter
  status = CStatus{};

  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request
  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=stat&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse out the status from the body
  if (!ParseStatusBody(sResponse.c_str(), status))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_STATUS_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseStatusBody(_In_ LPCSTR pszBody, _Out_ CStatus& status)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Initialize the output parameter
  status = CStatus{};

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then parse the status info line by line
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    bool bCurrentlyParsingCategories{false};
    bool bCurrentlyParsingSites{false};

    //Skip over the current line
    LPCSTR pszLine{pszBody};
#pragma warning(suppress: 26489)
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //iterate across all lines
    while (pszLine != nullptr)
    {
      //parse the current line and get the next line
#pragma warning(suppress: 26489)
      GetNextLine(pszLine, sLine);

      //Parse each line
      StringA sField;
#pragma warning(suppress: 26489)
      StringA sValue;
      if (bCurrentlyParsingCategories)
      {
        //Parse the category info
        if (sLine.substr(0, 1) == " ")
        {
#pragma warning(suppress: 26489)
          if (SplitSemiColonLine(sLine, sField, sValue))
          {
            status.m_Categories.push_back(MapFromProtocolString(sField));
            const int nValue{atoi(sValue.c_str())};
            status.m_CategoryEntries.push_back(nValue);
          }
        }
        else
          bCurrentlyParsingCategories = false;
      }

      if (bCurrentlyParsingSites)
      {
        //Parse the pending sites info
#pragma warning(suppress: 26489)
        if (sLine.substr(0, 1) == " ")
        {
#pragma warning(suppress: 26489)
          if (SplitSemiColonLine(sLine, sField, sValue))
          {
            status.m_PendingSites.push_back(MapFromProtocolString(sField));
            const int nValue{atoi(sValue.c_str())};
            status.m_PendingEntries.push_back(nValue);
          }
        }
        else
          bCurrentlyParsingSites = false;
      }

#pragma warning(suppress: 26489)
      if (sLine.find("Database entries by category:") == 0)
        bCurrentlyParsingCategories = true;
#pragma warning(suppress: 26489)
      else if (sLine.find("Pending File transmissions:") == 0)
        bCurrentlyParsingSites = true;
#pragma warning(suppress: 26489)
      else if (SplitSemiColonLine(sLine, sField, sValue))
      {
        if (sField == "current proto")
          status.m_nCurrentProtocol = atoi(sValue.c_str());
        else if (sField == "max proto")
          status.m_nMaxProtocol = atoi(sValue.c_str());
        else if (sField == "gets")
        {
          if (sValue == "yes")
            status.m_bGetsAllowed = true;
          else if (sValue == "no")
            status.m_bGetsAllowed = false;
          else
          {
            ATLTRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid gets response to the stat command\n"));
            SetLastError(ERROR_INVALID_DATA);
            return false;
          }
        }
        else if (sField == "updates")
        {
          if (sValue == "yes")
            status.m_bUpdatesAllowed = true;
          else if (sValue == "no")
            status.m_bUpdatesAllowed = false;
          else
          {
            ATLTRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid updates response to the stat command\n"));
            SetLastError(ERROR_INVALID_DATA);
            return false;
          }
        }
        else if (sField == "posting")
        {
          if (sValue == "yes")
            status.m_bPostingAllowed = true;
          else if (sValue == "no")
            status.m_bPostingAllowed = false;
          else
          {
            ATLTRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid posting response to the stat command\n"));
            SetLastError(ERROR_INVALID_DATA);
            return false;
          }
        }
        else if (sField == "quotes")
        {
          if (sValue == "yes")
            status.m_bQuotes = true;
          else if (sValue == "no")
            status.m_bQuotes = false;
          else
          {
            ATLTRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid quotes response to the stat command\n"));
            SetLastError(ERROR_INVALID_DATA);
            return false;
          }
        }
        else if (sField == "current users")
          status.m_nCurrentUsers = atoi(sValue.c_str());
        else if (sField == "max users")
          status.m_nMaxUsers = atoi(sValue.c_str());
        else if (sField == "strip ext")
        {
          if (sValue == "yes")
            status.m_bStripExtended = true;
          else if (sValue == "no")
            status.m_bStripExtended = false;
          else
          {
            ATLTRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid strip extended response to the stat command\n"));
            SetLastError(ERROR_INVALID_DATA);
            return false;
          }
        }
        else if (sField == "Database entries")
          status.m_nDatabaseEntries = atoi(sValue.c_str());
      }
    }

    //Everything went ok, so set the return value
    bSuccess = true;
  }

  return bSuccess;
}

void CDDB::CClient::Query(_In_ const CDDB::CSite& server, _In_ DWORD dwDiscID, _In_ const std::vector<CTrackPosition>& tracks, _Out_ std::vector<CQueryResult>& results)
{
  //Initialize the output parameter
  results.clear();

  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request

  String sDiscID;
#ifdef _UNICODE
  std::wostringstream ss;
#else
  std::ostringstream ss;
#endif //#ifdef _UNICODE
  ss << std::nouppercase << std::setfill(_T('0')) << std::setw(8) << std::hex << dwDiscID;
  sDiscID = ss.str();

  String sTracks;
#pragma warning(suppress: 26472)
  const int nTracks{ static_cast<int>(tracks.size()) - 1 }; //Number of tracks is 1 less than the size of the array as it also contains the lead out position
  ss.str(String{});
  ss << std::dec << nTracks;
  sTracks = ss.str();

  String sOffsets;
  ss.str(String{});
  for (int i{0}; i<nTracks; i++)
  {
    //Accumulate the frame number into the string
#pragma warning(suppress: 26446)
    const int nTrackLen{(tracks[i].m_nMinute * 60 * 75) + (tracks[i].m_nSecond * 75) + tracks[i].m_nFrame};
    ss << std::dec << nTrackLen;

    //Add a space between each offset
    if (i < (nTracks - 1))
      ss << _T("+");
  }
  sOffsets = ss.str();

  //Compute the playing length in seconds by using the lead out position
#pragma warning(suppress: 26446)
  const int nCDlength{(tracks[nTracks].m_nMinute * 60) + tracks[nTracks].m_nSecond};
  String sLength;
  ss.str(String{});
  ss << std::dec << nCDlength;
  sLength = ss.str();

  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=cddb+query+");
  sRequest += sDiscID;
  sRequest += _T("+");
  sRequest += sTracks;
  sRequest += _T("+");
  sRequest += sOffsets;
  sRequest += _T("+");
  sRequest += sLength;
  sRequest += _T("&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse out the query info from the body
  if (!ParseQueryBody(sResponse.c_str(), results))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_QUERY_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseQueryBody(_In_ LPCSTR pszBody, _Inout_ std::vector<CQueryResult>& results)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then parse the query data line by line
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    //Skip over the current line
    LPCSTR pszLine{pszBody};
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line
    while (pszLine != nullptr)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //If we can parse the current line then add it 
      //to the query array
      CQueryResult result;
#pragma warning(suppress: 26486)
      if (ParseQueryBodyHelper(sLine, result))
      {
        results.push_back(result);
        bSuccess = true;
      }
    }
  }
  else if ((nResponseCode == 200) && (strlen(pszBody) >= 5))
  {
    //found an exact match, If we can parse the 
    //current line then add it to the query array
    CQueryResult result;
#pragma warning(suppress: 26481 26486)
    if (ParseQueryBodyHelper(pszBody + 4, result))
    {
      results.push_back(result);
      bSuccess = true;
    }
  }
  else if (nResponseCode == 202)
  {
    //Receiving a 202 is not considered an error
    bSuccess = true;
  }

  return bSuccess;
}

void CDDB::CClient::Read(_In_ const CSite& server, _In_ DWORD dwDiscID, _In_ const String& sCategory, _Out_ CRecord& record)
{
  //Initialize the output parameter
  record = CRecord{};

  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request
  String sDiscID;
#ifdef _UNICODE
  std::wostringstream ss;
#else
  std::ostringstream ss;
#endif //#ifdef _UNICODE
  ss << std::nouppercase << std::setfill(_T('0')) << std::setw(8) << std::hex << dwDiscID;
  sDiscID = ss.str();
  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=cddb+read+");
  sRequest += sCategory;
  sRequest += _T("+");
  sRequest += sDiscID;
  sRequest += _T("&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse the CDDB record
  if (!ParseRecordBody(sResponse.c_str(), record))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_READ_RESPONSE, sResponse.c_str());

  if (record.m_dwDiscID != dwDiscID)
  {
    ATLTRACE(_T("CCDDB::Read, Returned DiscID from CDDB read (%08x) is different than the computed value (%08x), using the computed value\n"), record.m_dwDiscID, dwDiscID);
    record.m_dwDiscID = dwDiscID;
  }
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseRecordBody(_In_ LPCSTR pszBody, _Out_ CRecord& record)
{
  //initialize the output parameter
  record = CRecord{};

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then parse the query data line by line
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    //Skip over the current line
    LPCSTR pszLine{pszBody};
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //Hand of to the helper function
    if (!ParseRecordBodyHelper(pszLine, record))
      return false;
  }

  //Return the status code
  return true;
}

#ifdef _UNICODE
void CDDB::CClient::StringReplaceAll(_Inout_ String& sString, _In_ const String& sFrom, _In_ const String& sTo)
{
  size_t nPos{0};
  while ((nPos = sString.find(sFrom, nPos)) != String::npos)
  {
    sString.replace(nPos, sFrom.length(), sTo);
    nPos += sTo.length();
  }
}

void CDDB::CClient::StringTrim(_Inout_ String& sString)
{
  String::size_type nPos{sString.find_last_not_of(_T(" \n\r\t"))};
  if (nPos != String::npos)
    sString.resize(nPos + 1);

  nPos = sString.find_first_not_of(_T(" \n\r\t"));
  if (nPos != String::npos)
#pragma warning(suppress: 26489)
    sString = sString.substr(nPos);
}
#endif //#ifdef _UNICODE

void CDDB::CClient::StringReplaceAll(_Inout_ StringA& sString, _In_ const StringA& sFrom, _In_ const StringA& sTo)
{
  size_t nPos{0};
  while ((nPos = sString.find(sFrom, nPos)) != String::npos)
  {
    sString.replace(nPos, sFrom.length(), sTo);
    nPos += sTo.length();
  }
}

void CDDB::CClient::StringTrim(_Inout_ StringA& sString)
{
  StringA::size_type nPos{sString.find_last_not_of(" \n\r\t")};
  if (nPos != StringA::npos)
    sString.resize(nPos + 1);

  nPos = sString.find_first_not_of(" \n\r\t");
  if (nPos != StringA::npos)
#pragma warning(suppress: 26489)
    sString = sString.substr(nPos);
}

void CDDB::CClient::PrepareTextForSubmit(_Inout_ String& sText)
{
  StringReplaceAll(sText, _T("\\"), _T("\\\\"));
  StringReplaceAll(sText, _T("\n"), _T("\\n"));
#pragma warning(suppress: 26486 26489)
  StringReplaceAll(sText, _T("\t"), _T("\\t"));
}

CDDB::String CDDB::CClient::FormSubmitLine(_In_ const String& sKeyword, _In_ const String& sValue)
{
  String sLine;
  String sLocalValue{sValue};
  PrepareTextForSubmit(sLocalValue);

  do
  {
    const String::size_type nTotalSize{sKeyword.length() + sLocalValue.length() + 1};
    const String::size_type nLineLength{min(254, nTotalSize)};
    const String::size_type nValueLength{nLineLength - sKeyword.length() - 1};

    //Add to what will be the return value
    sLine += sKeyword;
    sLine += _T("=");
    sLine += sLocalValue.substr(0, nValueLength);
    sLine += _T("\r\n");

    //Prepare for the next time around
#pragma warning(suppress: 26489)
    sLocalValue = sLocalValue.substr(nValueLength);
  }
  while (sLocalValue.length());

  return sLine;
}

void CDDB::CClient::Submit(_In_ const CDDB::CSite& server, _In_ const String& sCategory, _In_ const String& sEmailAddress,
                           _In_ const CDDB::CRecord& record, _In_ const String& sRejectionNotice, _In_ bool bReallySubmit)
{
  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Form the request
  String sRequest(server.m_sAddress);
  hr = CreateRequest(sRequest.c_str(), L"POST");
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Form the headers
  std::wstring sHeaders(L"Category: ");
  sHeaders += ATL::CT2W(sCategory.c_str()).operator LPWSTR();
  sHeaders += L"\r\n";
  std::wostringstream ss2;
  ss2 << L"Discid: " << std::nouppercase << std::setfill(L'0') << std::setw(8) << std::hex << record.m_dwDiscID << L"\r\n";
  ss2 << L"User-Email: " << ATL::CT2W(sEmailAddress.c_str()).operator LPWSTR() << L"\r\n";
  sHeaders += ss2.str();
  if (bReallySubmit) //Then the submit mode
    sHeaders += L"Submit-Mode: submit\r\n";
  else
    sHeaders += L"Submit-Mode: test\r\n";
#ifdef _UNICODE
  if (m_nProtocolVersion >= 6)
    sHeaders += L"Charset: UTF-8\r\n";
  else
    sHeaders += L"Charset: ISO-8859-1\r\n";
#else
  sHeaders += L"Charset: ISO-8859-1\r\n";
#endif //#ifdef _UNICODE
#pragma warning(suppress: 26477)
  ATLASSERT(sRejectionNotice.length() <= 70); //Rejection notice is limited to 70 characters
  sHeaders += L"X-Cddbd-Note: "; //Then the rejection notice
  sHeaders += ATL::CT2W(sRejectionNotice.c_str()).operator LPWSTR();

  //Create the body for the request
  String sBody(_T("# xmcd\r\n#\r\n# Track frame offsets:\r\n"));
#ifdef _UNICODE
  std::wostringstream ss;
#else
  std::ostringstream ss;
#endif //#ifdef _UNICODE
  for (const auto& trackOffset : record.m_TrackOffsets)
  {
    ss.str(String{});
#pragma warning(suppress: 26446)
    ss << _T("# ") << trackOffset << _T("\r\n");
    sBody += ss.str();
  }
  ss.str(String{});
  ss << _T("#\r\n# Disc length: ") << record.m_nDiskLength << _T(" seconds\r\n");
  sBody += ss.str();
  ss.str(String{});
  ss << _T("#\r\n# Revision: ") << record.m_nDatabaseRevision << _T("\r\n");
  sBody += ss.str();
  String sBuf;
  sBuf = _T("# Submitted via : ");
  sBuf += record.m_sClientName;
  sBuf += _T(" ");
  sBuf += record.m_sClientVersion;
  sBuf += _T(" ");
  sBuf += record.m_sClientComments;
  PrepareTextForSubmit(sBuf);
#pragma warning(suppress: 26477)
  ATLASSERT(sBuf.length() <= 254); //Only 254 chars per line supported for the "# Submitted via:" line
  sBody += sBuf;
  sBody += _T("\r\n#\r\n");
  ss.str(String{});
  ss << _T("DISCID=") << std::nouppercase << std::setfill(_T('0')) << std::setw(8) << std::hex << record.m_dwDiscID << _T("\r\n");
  sBody += ss.str();
  if (record.m_sArtist.length())
  {
    sBuf = record.m_sArtist;
    sBuf += _T(" / ");
    sBuf += record.m_sTitle;
  }
  else
    sBuf = record.m_sTitle;
  sBuf = FormSubmitLine(_T("DTITLE"), sBuf);
  sBody += sBuf;
  if (m_nProtocolVersion >= 5)
  {
#pragma warning(suppress: 26477)
    ATLASSERT((record.m_nYear >= 0) && (record.m_nYear <= 9999));
    if (record.m_nYear != 0)
    {
      ss.str(String{});
      ss << _T("DYEAR=") << std::setfill(_T('0')) << std::setw(4) << std::dec << record.m_nYear << _T("\r\n");
      sBody += ss.str();
    }
    else
      sBody += _T("DYEAR=\r\n");
    sBuf = FormSubmitLine(_T("DGENRE"), record.m_sGenre);
    sBody += sBuf;
  }
  for (size_t i{0}; i<record.m_TrackTitles.size(); i++)
  {
    String sKeyword;
    ss.str(String{});
    ss << _T("TTITLE") << std::dec << i;
    sKeyword = ss.str();
#pragma warning(suppress: 26446)
    sBuf = FormSubmitLine(sKeyword, record.m_TrackTitles[i]);
    sBody += sBuf;
  }
  sBuf = FormSubmitLine(_T("EXTD"), record.m_ExtendedData);
  sBody += sBuf;
  for (size_t i{0}; i<record.m_ExtendedTrackData.size(); i++)
  {
    String sKeyword;
    ss.str(String{});
    ss << _T("EXTT") << std::dec << i;
    sKeyword = ss.str();
#pragma warning(suppress: 26446)
    sBuf = FormSubmitLine(sKeyword, record.m_ExtendedTrackData[i]);
    sBody += sBuf;
  }
  sBody += _T("PLAYORDER=");
  StringA sASCIIBody{MapToProtocolString(sBody)};

  //Read the response
  StringA sResponse;
#pragma warning(suppress: 26472)
  hr = ReadResponse(sASCIIBody.data(), static_cast<DWORD>(sASCIIBody.length()), sHeaders, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse out the response
  if (!ParseSubmitBody(sResponse.c_str()))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SUBMIT_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseSubmitBody(_In_ LPCSTR pszBody)
{
  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (20x) then return success
  return ((nResponseCode >= 200) && (nResponseCode <= 209));
}

void CDDB::CClient::MessageOfTheDay(_Inout_ String& sMessage, _In_ const String& sServer, _In_ const String& sAddress, _In_ INTERNET_PORT nPort)
{
  //Form the site which we will be getting the message from
  CSite server;
#pragma warning(suppress: 26489)
  server.m_sSite = sServer;
  server.m_nPort = nPort;
#pragma warning(suppress: 26489)
  server.m_sAddress = sAddress;

  //Let the other version 
#pragma warning(suppress: 26486)
  MessageOfTheDay(server, sMessage);
}

void CDDB::CClient::MessageOfTheDay(_In_ const CSite& server, _Inout_ String& sMessage)
{
  //Create the internet session
  HRESULT hr{CreateSession()};
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create the internet connection
  hr = CreateConnection(server.m_sSite.c_str(), server.m_nPort);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Create and issue the request
  String sRequest(server.m_sAddress);
  sRequest += _T("?cmd=motd&");
  sRequest += GetHelloCommand();
  hr = CreateRequest(sRequest.c_str(), nullptr);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Read the response
  StringA sResponse;
  hr = ReadResponse(nullptr, 0, WinHTTPWrappers::String{}, sResponse);
  if (FAILED(hr))
    ThrowCDDBException(hr);

  //Parse out the message from the body
  if (!ParseMessageBody(sResponse.c_str(), sMessage))
    ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_MOTD_RESPONSE, sResponse.c_str());
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseMessageBody(_In_ LPCSTR pszBody, _Inout_ String& sMessage)
{
  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  //Initialize the output parameter
  sMessage.clear();

  //From the HTTP body get the CDDB response code
  const int nResponseCode{GetCDDBResponseCode(pszBody)};

  //If the response is in the expected range (21x) then copy the message into "sMessage"
  if ((nResponseCode >= 210) && (nResponseCode <= 219))
  {
    //Skip over the current line
    LPCSTR pszLine{pszBody};
    StringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line
    while (pszLine != nullptr)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      if (sLine != ".")
      {
#pragma warning(suppress: 26489)
        sMessage += ATL::CA2T{sLine.c_str()};
#pragma warning(suppress: 26489)
        sMessage += _T("\r\n");
      }

      bSuccess = true;
    }
  }

  return bSuccess;
}

CDDB::String CDDB::CClient::MapFromProtocolString(_In_ const StringA& sInput)
{
  //What will be the return value from this function
  String sOutput;

#ifdef _UNICODE
  if (m_nProtocolVersion >= 6)
    sOutput = ATL::CA2W{sInput.c_str(), CP_UTF8};
  else
    sOutput = ATL::CA2W{sInput.c_str()}; //Do a simple conversion from ASCII to Unicode
#else
  sOutput = sInput;
#endif //#ifdef _UNICODE

  return sOutput;
}

CDDB::StringA CDDB::CClient::MapToProtocolString(_In_ const String& sInput)
{
  //What will be the return value from this function
  StringA sOutput;

#ifdef _UNICODE
  if (m_nProtocolVersion >= 6)
    sOutput = ATL::CW2A{sInput.c_str(), CP_UTF8};
  else
    sOutput = ATL::CW2A{sInput.c_str()}; //Do a simple conversion from Unicode to ASCII
#else
  sOutput = sInput;
#endif //#ifdef _UNICODE

  return sOutput;
}

void CDDB::CClient::GetNextLine(_Inout_ LPCSTR& pszLine, _Inout_ StringA& sLine)
{
  if (pszLine != nullptr)
  {
    //Find the start of the next line. Both DOS/Windows (\r\n)
    //and UNIX (\n) terminators are valid. First try to
    //find a DOS EOL
    const char* pszNextLine{strstr(pszLine, "\r\n")};
    if (pszNextLine != nullptr)
    {
      sLine.assign(pszLine, pszNextLine - pszLine);
#pragma warning(suppress: 26481)
      pszLine = pszNextLine + 2;
    }
    else
    {
      //Failed to get a DOS/Windows EOL, try to get a UNIX EOL instead
      pszNextLine = strstr(pszLine, "\n");
      if (pszNextLine != nullptr)
      {
        sLine.assign(pszLine, pszNextLine - pszLine);
#pragma warning(suppress: 26481)
        pszLine = pszNextLine + 1;
      }
      else
        pszLine = nullptr;
    }
  }
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseKeywordLine(_In_ const StringA& sLine, _Out_ StringA& sKeyword, _Out_ StringA& sValue)
{
  //what will be the return value from this function (assume the worst)
  bool bSuccess{false};

  const auto nSeparator{sLine.find("=")};
  if (nSeparator != StringA::npos)
  {
    bSuccess = true;

    sKeyword = sLine.substr(0, nSeparator);
    sValue = sLine.substr(nSeparator + 1);

    //Replace all strings as specified in the cddb document
    StringReplaceAll(sValue, "\\n", "\n");
#pragma warning(suppress: 26486 26489)
    StringReplaceAll(sValue, "\\t", "\t");
#pragma warning(suppress: 26486 26489)
    StringReplaceAll(sValue, "\\\\", "\\");
  }

  return bSuccess;
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseCommentLine(_In_ const StringA& sLine, _Out_ StringA& sValue)
{
  //What will be the return value from the function (assume the worst)
  bool bSuccess{false};

  //Do the actual parsing
  if (sLine.substr(0, 1) == "#")
  {
    bSuccess = true;
    sValue = sLine.substr(1);
    StringTrim(sValue);
  }

  return bSuccess;
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseRecordBodyHelper(_In_ const StringA& sRecord, _Out_ CRecord& record)
{
  //Initialize the output parameter
  record = CRecord{};

  //Setup all the booleans which indicate a successful record parse.
  //These will be checked at the end of the function
  bool bFoundXMCDComment{false};
  bool bParsingTrackFrameOffsets{false};
  bool bFoundDiscLength{false};
  bool bFoundSubmittedVia{false};
  bool bFoundDISCID{false};
  bool bFoundDTITLE{false};
#pragma warning(suppress: 26489)
  String sTitle;
  String sPlayOrder;

  //Loop thro each line 
  LPCSTR pszCurLine{sRecord.c_str()};
  while (pszCurLine != nullptr)
  {
    //Get the next line before we parse the current one
    StringA sLine;
#pragma warning(suppress: 26486)
    GetNextLine(pszCurLine, sLine);

    //Is it a comment line?
#pragma warning(suppress: 26489)
    StringA sValue;
    if (ParseCommentLine(sLine, sValue))
    {
      if (sValue.find("xmcd") == 0)
      {
        //Handle the "# xmcd" line
        bFoundXMCDComment = true;
      }
      else if (sValue.find("Track frame offsets:") != StringA::npos)
      {
        //Handle the "# Track frame offsets:" line
        bParsingTrackFrameOffsets = true;
      }
      else if (sValue.find("Disc length:") == 0)
      {
        //If we've found the disc length line then there are no more track frame offset lines
        bParsingTrackFrameOffsets = false;

        //Handle the "# Disc length:" line
        bFoundDiscLength = true;
#pragma warning(suppress: 26489)
        sValue = sValue.substr(13);
        record.m_nDiskLength = atoi(sValue.c_str());
      }
      else if (sValue.find("Revision:") == 0)
      {
        //Handle the "# Revision:" line
#pragma warning(suppress: 26489)
        sValue = sValue.substr(9);
        record.m_nDatabaseRevision = atoi(sValue.c_str());
      }
      else if (sValue.find("Submitted via:") == 0)
      {
        //Handle the "# Submitted via:" line
        bFoundSubmittedVia = true;

        static const char* pszSeparators{" "};
#pragma warning(suppress: 26489)
        sValue = sValue.substr(14);
#pragma warning(suppress: 26446)
        char* pszValue{sValue.data()};
        char* pszNextToken{nullptr};
        char* pszToken{strtok_s(pszValue, pszSeparators, &pszNextToken)};
        if (pszToken != nullptr)
          record.m_sClientName = MapFromProtocolString(pszToken);
        else
        {
          ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a valid Submitted via: line\n"));
          SetLastError(ERROR_INVALID_DATA);
          return false;
        }

        //Pull out the client version
        pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
        if (pszToken != nullptr)
          record.m_sClientVersion = MapFromProtocolString(pszToken);
        else
        {
          ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a valid Submitted via: line\n"));
          SetLastError(ERROR_INVALID_DATA);
          return false;
        }

        //Pull out the client comments
        pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
        if (pszToken != nullptr)
          record.m_sClientComments = MapFromProtocolString(pszToken);
      }
      else if (bParsingTrackFrameOffsets)
      {
        if (sValue.length())
        {
          //Handle all the actual track frame offsets lines
          const int nTrackOffset{atoi(sValue.c_str())};
          record.m_TrackOffsets.push_back(nTrackOffset);
        }
      }
    }
    else
    {
      //Must not be a comment line
      StringA sKeyword;
      if (ParseKeywordLine(sLine, sKeyword, sValue))
      {
        if (sKeyword == "DISCID")
        {
          if (sValue.length() >= 8)
          {
            //handle the "DISCID" line
            bFoundDISCID = true;
            sscanf_s(sValue.c_str(), "%x", &record.m_dwDiscID);
          }
        }
        else if (sKeyword == "DTITLE")
        {
          //Handle the "DTITLE" line
          bFoundDTITLE = true;
          sTitle += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "DYEAR")
        {
          //Handle the "DYEAR" line
          record.m_nYear = atoi(sValue.c_str());
        }
        else if (sKeyword == "DGENRE")
        {
          //Handle the "DGENRE" line
          record.m_sGenre += MapFromProtocolString(sValue);
        }
        else if (sKeyword.find("TTITLE") == 0)
        {
          //Handle the "TTITLE" lines

          //work out the offset and add the string
          StringA sTrack{sKeyword.substr(6)};
#pragma warning(suppress: 26472)
          const auto nOffset{static_cast<size_t>(atoi(sTrack.c_str()))};
          record.m_TrackTitles.resize(nOffset + 1);
#pragma warning(suppress: 26446)
          record.m_TrackTitles[nOffset] += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "EXTD")
        {
          //Handle any "EXTD" lines
          record.m_ExtendedData += MapFromProtocolString(sValue);
        }
        else if (sKeyword.find("EXTT") == 0)
        {
          //work out the offset and add the string
          StringA sTrack{sKeyword.substr(4)};
#pragma warning(suppress: 26472)
          const auto nOffset{static_cast<size_t>(atoi(sTrack.c_str()))};
          record.m_ExtendedTrackData.resize(nOffset + 1);
#pragma warning(suppress: 26446)
          record.m_ExtendedTrackData[nOffset] += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "PLAYORDER")
        {
          //Handle the "PLAYORDER" line
          sPlayOrder += MapFromProtocolString(sValue);
        }
      }
    }
  }

  //Handle all the parsing errors
  if (!bFoundXMCDComment)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a XMCD comment line\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (record.m_TrackOffsets.size() == 0)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with any Track frame offset lines\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (!bFoundDiscLength)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a Disc length line\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (!bFoundSubmittedVia)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a Submitted via line\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (!bFoundDISCID)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a DISCID line\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (!bFoundDTITLE)
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a DTITLE line\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (record.m_TrackTitles.size() != record.m_ExtendedTrackData.size())
  {
    ATLTRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with an equal amount of TTITLE lines to EXTT lines\n"));
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //Put together the details from the "DTITLE" line(s)
  String::size_type nFind{sTitle.find(_T(" / "))};
  if (nFind != String::npos)
  {
    record.m_sArtist = sTitle.substr(0, nFind);
#pragma warning(suppress: 26489)
    record.m_sTitle = sTitle.substr(nFind + 3);
  }
  else
  {
    //Artist & Title are the same
    record.m_sTitle = sTitle;
    record.m_sArtist = sTitle;
  }

  //Handle the "PLAYORDER" line
  bool bContinue{true};
  do
  {
    nFind = sPlayOrder.find(_T(","));
    if (nFind != String::npos)
    {
#pragma warning(suppress: 26489)
      String sNum{sPlayOrder.substr(0, nFind)};
      const int nNum{_ttoi(sNum.c_str())};
      record.m_PlayOrder.push_back(nNum);
#pragma warning(suppress: 26489)
      sPlayOrder = sPlayOrder.substr(nFind + 1);
      bContinue = true;
    }
    else
    {
      if (sPlayOrder.length())
      {
        const int nNum{_ttoi(sPlayOrder.c_str())};
        record.m_PlayOrder.push_back(nNum);
      }
      bContinue = false;
    }
  }
  while (bContinue);

  return true;
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseSiteBodyHelper(_In_ const StringA& sLine, _Out_ CSite& site)
{
  //Initialize the output parameter
  site = CSite{};

  //Create a local copy of sLine so that we can use strtok_s on it
  StringA sLocalLine{sLine};

  //Pull out "m_sSite"
  static const char* pszSeparators{" "};
#pragma warning(suppress: 26446)
  char* pszValue{sLocalLine.data()};
  char* pszNextToken{nullptr};
  LPSTR pszToken{strtok_s(pszValue, pszSeparators, &pszNextToken)};
  if (pszToken != nullptr)
    site.m_sSite = MapFromProtocolString(pszToken);
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (m_nProtocolVersion >= 3)
  {
    //Pull out "m_sProtocol"
    pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
    if (pszToken != nullptr)
      site.m_sProtocol = MapFromProtocolString(pszToken);
    else
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }
  }

  //Pull out "m_nPort"
  pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
  if (pszToken != nullptr)
#pragma warning(suppress: 26472)
    site.m_nPort = static_cast<INTERNET_PORT>(atoi(pszToken));
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  if (m_nProtocolVersion >= 3)
  {
    //Pull out "m_sAddress"
    pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
    if (pszToken != nullptr)
      site.m_sAddress = MapFromProtocolString(pszToken);
    else
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }
  }

  //Pull out the latitude values
  pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
  if (pszToken != nullptr)
  {
    String sLatitude{MapFromProtocolString(pszToken)};
  
    //Latitude field must be 7 characters long
    if (sLatitude.length() != 7)
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }

    //Parse out the northing field
#pragma warning(suppress: 26446)
    if (sLatitude[0] == _T('N'))
      site.m_bNorthing = true;
#pragma warning(suppress: 26446)
    else if (sLatitude[0] == _T('S'))
      site.m_bNorthing = false;
    else
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }

    //Parse out the Latitude degrees
#pragma warning(suppress: 26446)
    String sNum{&(sLatitude[1]), 3};
    const int nLatDeg{_ttoi(sNum.c_str())};

    //Parse out the Latitude minutes
#pragma warning(suppress: 26446)
    sNum = String{&(sLatitude[5]), 2};
    const int nLatMin{_ttoi(sNum.c_str())};
    site.m_nLatitudeMinutes = nLatMin + 60*nLatDeg;
  }
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //Pull out the longitude values
  pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
  if (pszToken != nullptr)
  {
    String sLongitude{MapFromProtocolString(pszToken)};

    //Latitude field must be 7 characters long
    if (sLongitude.length() != 7)
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }

    //Parse out the northing field
#pragma warning(suppress: 26446)
    if (sLongitude[0] == _T('E'))
      site.m_bEasting = true;
#pragma warning(suppress: 26446)
    else if (sLongitude[0] == _T('W'))
      site.m_bEasting = false;
    else
    {
      SetLastError(ERROR_INVALID_DATA);
      return false;
    }

    //Parse out the Longitude degrees
#pragma warning(suppress: 26446)
    String sNum{&(sLongitude[1]), 3};
    const int nLongDeg{_ttoi(sNum.c_str())};

    //Parse out the Longitude minutes
#pragma warning(suppress: 26446)
    sNum = String{&(sLongitude[5]), 2};
    const int nLongMin{_ttoi(sNum.c_str())};
    site.m_nLongitudeMinutes = nLongMin + 60*nLongDeg;
  }
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //What's left is put into the description field
  if (pszNextToken != nullptr)
    site.m_sDescription = MapFromProtocolString(pszNextToken);

  //Everything has been parsed correctly
  return true;
}

_Return_type_success_(return != 0) bool CDDB::CClient::ParseQueryBodyHelper(_In_ const StringA& sLine, _Out_ CQueryResult& result)
{
  //Initialize the output parameter
  result = CQueryResult{};

  //Create a local copy of sLine so that we can use strtok_s on it
  StringA sLocalLine{sLine};

  //Pull out the category
  static const char* pszSeparators{" "};
  char* pszValue{sLocalLine.data()};
  char* pszNextToken{nullptr};
  LPSTR pszToken{strtok_s(pszValue, pszSeparators, &pszNextToken)};
  if (pszToken != nullptr)
    result.m_sCategory = MapFromProtocolString(pszToken);
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //Pull out the CDDB DISCID
  pszToken = strtok_s(nullptr, pszSeparators, &pszNextToken);
  if (pszToken != nullptr)
    sscanf_s(pszToken, "%x", &result.m_dwDiscID);
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //Pull out the Artist and Title
  if (pszNextToken != nullptr)
    result.m_sTitle = MapFromProtocolString(pszNextToken);
  else
  {
    SetLastError(ERROR_INVALID_DATA);
    return false;
  }

  //Remove any EOL if it is on the end
  String::size_type nEOLPos{result.m_sTitle.find(_T("\r\n"))};
  if (nEOLPos == String::npos)
    nEOLPos = result.m_sTitle.find(_T("\n"));
  if (nEOLPos != String::npos)
#pragma warning(suppress: 26489)
    result.m_sTitle = result.m_sTitle.substr(0, nEOLPos);

  //Split into the artist and title
  const auto nSlashPos{result.m_sTitle.find(_T('/'))};
  if (nSlashPos != String::npos)
  {
    result.m_sArtist = result.m_sTitle.substr(0, nSlashPos);
#pragma warning(suppress: 26489)
    result.m_sTitle = result.m_sTitle.substr(nSlashPos + 1);

    //Remove any trailing or leading spaces
    StringTrim(result.m_sArtist);
    StringTrim(result.m_sTitle);
  }

  return true;
}

void CDDB::CClient::ThrowCDDBException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_z_ LPCSTR pszLastResponse)
{
  if (dwError == 0)
    dwError = GetLastError();
  CException e{dwError, dwFacility, pszLastResponse};
  ATLTRACE(_T("Warning: throwing CDDB::CException for error 0x%08X\n"), e.m_hr);
  throw e;
}

void CDDB::CClient::ThrowCDDBException(_In_ HRESULT hr, _In_z_ LPCSTR pszLastResponse)
{
  CException e{hr, pszLastResponse};
  ATLTRACE(_T("Warning: throwing CDDB::CException for error 0x%08X\n"), e.m_hr);
  throw e;
}

void CDDB::CClient::ThrowCDDBMMException(_In_ DWORD dwError, _In_z_ LPCSTR pszLastResponse)
{
  CException e{true, dwError, pszLastResponse};
  ATLTRACE(_T("Warning: throwing CDDB::CException for error 0x%08X\n"), e.m_hr);
  throw e;
}

#pragma warning(suppress: 26440)
HRESULT CDDB::CClient::CreateSession()
{
  if (m_Session == nullptr)
  {
#pragma warning(suppress: 26477)
    return m_Session.Initialize(m_sUserAgent.length() ? m_sUserAgent.c_str() : nullptr, WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0);
  }
  return S_OK;
}

HRESULT CDDB::CClient::CreateConnection(_In_z_ LPCTSTR pszServer, _In_ INTERNET_PORT nPort)
{
  if (m_Connection != nullptr)
    m_Connection.Close();
  return m_Connection.Initialize(m_Session, ATL::CT2W{pszServer}, nPort);
}

HRESULT CDDB::CClient::CreateRequest(_In_z_ LPCTSTR pszRequest, _In_opt_z_ LPCWSTR pwszVerb)
{
  if (m_Request != nullptr)
    m_Request.Close();
  return m_Request.Initialize(m_Connection, ATL::CT2W{pszRequest}, pwszVerb);
}

HRESULT CDDB::CClient::ReadResponse(_In_reads_bytes_opt_(dwRequestSize) LPVOID pRequest, _In_ DWORD dwRequestSize, _In_ const WinHTTPWrappers::String& sHeaders, _Inout_ StringA& sResponse)
{
  //Do the download
  const HRESULT hr{m_Request.SendRequestSync(pRequest, dwRequestSize, sHeaders)};
  if (SUCCEEDED(hr))
#pragma warning(suppress: 26490)
    sResponse.insert(0, reinterpret_cast<const char*>(m_Request.m_Response.data()), m_Request.m_Response.size());
  m_Request.ReleaseResources();
  if (FAILED(hr))
    ThrowCDDBException(hr, FACILITY_WIN32);
  return hr;
}

#pragma warning(suppress: 26440)
void CDDB::CClient::AttachSession(_In_ HINTERNET hSession)
{
  m_Session.Attach(hSession);
}

#pragma warning(suppress: 26440)
HINTERNET CDDB::CClient::DetachSession()
{
  return m_Session.Detach();
}
